import xbmcplugin

from resources.lib.api.api import API
from resources.lib.const import ROUTE, STREAM_AUTOSELECT, explicit_genres, \
    MEDIA_TYPE, COMMAND, DIR_TYPE, SETTINGS
from resources.lib.gui import MainMenuFolderItem, DirectoryItem, MediaItem, ParentItem
from resources.lib.gui.renderers.directory_renderer import DirectoryRenderer
from resources.lib.gui.renderers.media_info_renderer import MediaInfoRenderer
from resources.lib.kodilogging import logger
from resources.lib.routing.router import router
from resources.lib.storage.settings import settings
from resources.lib.utils import first
from resources.lib.utils.kodiutils import get_string

has_children_map = {
    False: MediaItem,
    True: ParentItem
}


class MediaListRenderer:
    def __init__(self):
        """
        :param callable on_stream_selected: Called when a stream is selected.
        """
        self.media_has_children_map = {
            True: lambda media: router.get_url_api(ROUTE.GET_MEDIA, API.FILTER.parent(media.get('_id'))),
            False: lambda media: MediaListRenderer._get_stream_url(media),
        }

    def __call__(self, handle, media_list_gui, pagination):
        logger.debug('Renderer %s call' % self)
        MediaListRenderer.add_navigation(media_list_gui, bottom=pagination)
        if pagination:
            MediaListRenderer.add_pagination(media_list_gui, pagination)
        MediaListRenderer.render(handle, DIR_TYPE.MOVIES, [item.build() for item in media_list_gui])

    def __repr__(self):
        return self.__class__.__name__

    @staticmethod
    def _get_stream_url(media):
        api_url = API.URL.streams(media.get('_id'))
        source = API.get_source(media)
        return router.get_url(ROUTE.GET_STREAMS,
                              url=api_url,
                              media_id=media.get('_id'),
                              root_parent_id=source.get('root_parent'))

    def build_media_list(self, media_list, title_builder=MediaInfoRenderer.TITLE.default, *args):
        media_list_gui = []
        for media in media_list.get('data'):
            source = API.get_source(media)
            info_labels, art = MediaInfoRenderer.merge_info_labels(source)
            title = title_builder(source, *args) or info_labels.get('originaltitle')
            cast = source.get('cast')
            has_children = source.get('children_count', 0) > 0
            has_streams = MediaInfoRenderer.stream_available(source)
            # if not has_children and not has_streams:
            #     continue
            url = self.media_has_children_map[has_children](media)
            gui_item = MediaListRenderer.build_media_item_gui(has_children_map[has_children], source, url,
                                                              title, has_streams, art, cast,
                                                              info_labels)
            media_list_gui.append(gui_item)

        return media_list_gui

    @staticmethod
    def render(handle, dir_type, list_items):
        # as_type=None causes all items to show their icons and spam bullshit in log
        with DirectoryRenderer.start_directory(handle, as_type=dir_type):
            xbmcplugin.addDirectoryItems(handle, list_items)

    @staticmethod
    def add_navigation(list_items, bottom=False):
        if settings.as_bool(SETTINGS.SHOW_TOP_NAVIGATION):
            item = MainMenuFolderItem(url=router.get_url_command(COMMAND.CLEAR_PATH))
            list_items.insert(0, item)
        if bottom:
            item = MainMenuFolderItem(url=router.get_url_command(COMMAND.CLEAR_PATH))
            list_items.append(item)

    @staticmethod
    def add_pagination(list_items, pagination):
        if settings.as_bool(SETTINGS.SHOW_TOP_NAVIGATION):
            list_items.insert(0, MediaListRenderer.next_page_item(pagination))
        list_items.append(MediaListRenderer.next_page_item(pagination))

    @staticmethod
    def next_page_item(pagination):
        return DirectoryItem(
            title=MediaListRenderer.next_page_title(pagination['page'], pagination['pageCount']),
            url=router.get_url_api(ROUTE.GET_MEDIA, pagination['next']))

    @staticmethod
    def next_page_title(page, page_count):
        return '{} ({}/{})'.format(get_string(30203), page, page_count)

    @staticmethod
    def _autoselect_stream(streams, quality, language, max_bitrate, avoid_video_codecs, channels):
        """
        Function takes streams to select and indexes of desired quality and language as defined in const.py.

        It enumerates all streams and try to find best match in this order (disqualifying all blacklisted streams):
        1. find stream with desired language/audio
        2. find stream within the bitrate limit
        3. find stream with desired quality

        If all criteria matches at least in one stream use that (if more streams qualifies, use the first one)

        If not all of them satisfy the selection
        1. try to find stream with language match (if more streams qualifies, use the first one)
        2. try to find stream within the bitrate limit
        3. try to find stream with quality match (if more streams qualifies, use the first one)

        If no stream qualifies, use the first avialable stream.

        :param list streams: list of available streams
        :param int quality: desired quality (index in STREAM_AUTOSELECT.QUALITIES)
        :param int language: language desired language (index in STREAM_AUTOSELECT.LANGUAGES)
        :param int max_bit_rate: desired maximal bitrate
        :param list avoid_video_codecs: list of codecs to disqualify from autoselect
        :param float channels: preferred audio channels configuration

        returns selected stream
        """
        selected = streams[0]
        reason = 'no-match'
        avoid_video_codecs = map(str.upper, avoid_video_codecs)
        try:
            with_right_audio = set([])
            with_right_video = set([])
            with_right_bitrate = set([])
            for i, stream in enumerate(streams):
                if stream.get('codec', '').upper() in avoid_video_codecs:
                    logger.debug('dropping stream because video codec %s is blacklisted in settings',
                                 stream.get('codec'))
                    continue
                for audio in stream['audio']:
                    if audio["language"] == STREAM_AUTOSELECT.LANGUAGES[language]:
                        with_right_audio.add(i)
                        break
                if stream['quality'] == STREAM_AUTOSELECT.QUALITIES[quality]:
                    with_right_video.add(i)
                if max_bitrate != 0 and MediaListRenderer._is_bitrate_in_limit(stream.get('bitrate', 0), max_bitrate):
                    with_right_bitrate.add(i)

            to_intersect = filter(
                lambda s: len(s) > 0,
                [with_right_audio, with_right_video, with_right_bitrate])

            intersection = reduce(lambda s1, s2: s1 & s2, to_intersect) if len(to_intersect) > 0 else set()

            if len(intersection) > 0:
                reason = 'right audio, right quality, right_bitrate'
                eligible = intersection
            elif len(with_right_audio) > 0:
                reason = 'right audio'
                eligible = with_right_audio
            elif len(with_right_bitrate) > 0:
                reason = 'right bitrate'
                eligible = with_right_bitrate
            elif len(with_right_video) > 0:
                reason = 'right video'
                eligible = with_right_video

            eligible_streams = map(lambda i: streams[i], eligible)
            selected = first(eligible_streams)

            if channels is not None:
                def filter_channels(stream):
                    return len(filter(lambda audio: audio['channels'] == channels, stream['audio'])) > 0

                with_channels = first(filter(filter_channels, eligible_streams))
                if with_channels is not None:
                    selected = with_channels
                    reason += ', right channels'

        except Exception as e:
            reason = 'error'
            logger.error("autoselect: failed to auto select stream: %s", e)

        logger.debug(
            "autoselect: auto selected stream: %s (reason: %s) (criteria: lang => %s, quality => %s, max_bitrate => %s, avoiding: %s, channels: %s)",
            selected, reason, STREAM_AUTOSELECT.LANGUAGES[language], STREAM_AUTOSELECT.QUALITIES[quality], max_bitrate,
            avoid_video_codecs, channels)

        return selected

    @staticmethod
    def _is_bitrate_in_limit(bitrate, limit, tolerance=100):
        """
        :param int bitrate: stream bitrate
        :param int lmit: max bitrate in Mbit/s
        :param int tolerance: tolerance in Kbits/s

        returns True if bitrate is in limit + given tolerance or limit is set to 0; False otherwise
        """

        if limit <= 0:
            return True

        limit = float(limit)
        try:
            bitrate = float(bitrate) / 1024 / 1024
        except:
            bitrate = 0
            limit = 0
        return bitrate <= limit + float(tolerance) / 1024

    @staticmethod
    def build_media_item_gui(item_type, source, url, title, playable, art=None, cast=None, info_labels=None):
        # STRINGS.TITLE_YEAR.format(title=info_labels.get('title').encode('utf-8'), year=info_labels.get('year')
        return item_type(
            title=title,
            url=url,
            art=art,
            cast=cast,
            playable=playable,
            info_labels=info_labels,
            stream_info=source.get('stream_info'),
            services=source.get('services')
        )

    @staticmethod
    def explicit_filter(media_list):
        filtered_list = []
        for media in media_list.get('data'):
            source = API.get_source(media)
            genres = source.get('info_labels').get('genre')
            is_blocked = bool(set(genres).intersection(explicit_genres))
            if is_blocked:
                continue
            filtered_list.append(media)
        media_list.update({'data': filtered_list})
